/*****************************************************************************
//Atom.h
//Created by Richard Shadrach
//2008-7-29
//
//Description: 
//Represents a single atom.  Equality of atoms is checked by
//comparing the atom number.  Thus if you want expected behavior, you should
//ensure that each atom has a unique number.  As of this writing, charge and
//type are not used, but are commonly provided with the input files.
//
//Definitions: 
//Neighbor: Any atom that is directly bonded to *this.
//Chunk: The largest collection of atoms around *this such that every atom
//       in the chunk is within distance 2 where distance is traversing a bond.
//R4 and R6: Represents the function used to calculate the Born Radii, either
//           1/(r^4) or 1/(r^6).  
//Analytic: Refers to the method of using pairwise descreening with the
//          analytic function to calcute the Born Radii.
//Type: Element name.
//Sum: The sum of the pairwise descreening method.  Note that it also refers
       to the equivalent value when using numerical integration over VDW or
       molecular volume.
//Center of Mass: The CoM for the chunk.
//
//Abbreviations:
//com: center of mass
//BR: Born Radius (typically means the inverse of the Born Radius)
//vdw: Van Der Waals volume

//SYSTEM  : Tested on Linux Ubuntu 1.0 g++ 4.1.1 (Boris Aguilar) 


*****************************************************************************/

#ifndef ATOM_H
#define ATOM_H

#include <iostream>
#include <vector>
#include <set>

using namespace std;

#include "Functions.h"

class Atom {
public:
	Atom(double paramX, double paramY, double paramZ, double paramRadius, 
		int paramNumber, double paramCharge = 0., string paramType = "?");
	Atom(const Atom& a);

	bool operator == (const Atom& a) const;
	friend ostream& operator << (ostream& out, const Atom& a);
	
	void CoM(double paramX, double paramY, double paramZ);
	void CoM(double &paramX, double &paramY, double &paramZ) const;

	double distance(const Atom& a) const;
	double distance(double paramX, double paramY, double paramZ) const;
	
	bool addAtomToNeighbors(int atomNumber);
	set<int> Neighbors();
	set<int>::iterator neighborsBegin();
	set<int>::iterator neighborsEnd();
	int neighborsSize();
	void resetNeighbors();
	bool isInNeighbors(int atomNumber);

	bool addAtomToChunk(int atomNumber);
	bool addGroupToChunk(const set<int> &group);
	bool removeAtomFromChunk(int atomNumber);
	set<int>::iterator chunkBegin();
	set<int>::iterator chunkEnd();
	bool isInChunk(int atomNumber);
	set<int> Chunk();
	int chunkSize() const;

	// No necks for this group of atoms
	bool addAtomToNoNeck(int atomNumber);
        bool addGroupToNoNeck(const set<int> &group);
        bool removeAtomFromNoNeck(int atomNumber);
        set<int>::iterator noneckBegin();
        set<int>::iterator noneckEnd();
        bool isInNoNeck(int atomNumber);
        set<int> NoNeck();
        int NoNeckSize() const;


	double X() const;
	double Y() const;
	double Z() const;
	double OrigRadius() const;
	double Radius() const;
	int Number() const;
	double Charge() const;
	string Type() const;	
	
	double AnalyticR6BR() const;
	double VdwR4Sum() const;
	double VdwR6Sum() const;
	double ChunkR6BR() const;
	double ChunkVolume() const;
	double MolecularR6Sum() const;
	double ChunkR6Packing() const;  // Obtain the chunk packing factor B.A

	void X(double paramX);
	void Y(double paramY);
	void Z(double paramZ);
	void Radius(double paramRadius);

	void AnalyticR6BR(double value);
	void VdwR4Sum(double value);
	void VdwR6Sum(double value);
	void ChunkR6BR(double value);
	void ChunkVolume(double value);
	void MolecularR6Sum(double value);
	void ChunkR6Packing(double value); //Set the chunk packing factor


        int Region;  // Define if Atom is in Chunk, or in No Neck or in VDW region
                     // 0 for VDW, 1 for NoNeck, and 2 for VDW.

        double sigma_i;   // Lennard Jones parameters for
	double epsilon_i; // the atom type
        int VDWindex; // index for the LennarJones table

        double qeff ; // effective charge Abhisheck model

private:
	double x, y, z;   //Coordinates of the position of the Atom
	double origRadius;  //Originial Radius of the Atom
	double radius;    // Possibly modief radious (by a rescaling factor) 
	int number;      // Numbre in the pqr file
	double charge;   // Charge of the Atom
	string type;	 // Ytpe of the atom, e.g H, C, O. CA, CB, .. etc
	
	set<int> neighbors; // list of Neigbors
	set<int> chunk;	   // List of atoms that form the chunk of the atom 
	
	set<int> noneck;   // List of atoms with no neck contribution
	
	double com_x, com_y, com_z; //
	double analyticR6BR;      // Analytically computed R6 of the atom 
	double vdwR4Sum;         //
	double vdwR6Sum;         //
	double chunkR6BR;       // R6 contribution of the chunk
	double chunkVolume;     // Volumen of the chunks
	double molecularR6Sum;     //

	double chunkR6Packing;  //Chunk Packing factor, added by B.A.
};

#endif
