#include "Functions.h"

Functions::Functions() {
}
	
double Functions::Compare(double LHS, double RHS, double epsilon) {
	if (fabs(LHS - RHS) <= epsilon * fabs((LHS - RHS) / 2)) return 0.0;
	return LHS - RHS;
}

double Functions::Max(double LHS, double RHS) {
	if (LHS > RHS) {
		return LHS;
	}
	return RHS;
}

double Functions::CubeRoot(double value) {
	if (value >= 0) return pow(value, 1/3.0);
	return -pow(-value, 1/3.0);
}

double Functions::Pow(double value, double power) {
	if (value >= 0) return pow(value, power);
	return -pow(-value, power);
}	

void Functions::Debug(const char* fmt, ...)
{
    assert(fmt && *fmt);

    #ifdef debug
        va_list ap;

        va_start(ap, fmt);
        fputs("DEBUG : ", stdout);
        vfprintf(stdout, fmt, ap);
        fflush(stdout);
        va_end(ap);
    #endif
}

bool Functions::equal(double u, double v, double epsilon) {
  return (u + epsilon > v) && (u - epsilon < v);
}


//////////////////////////////////////////////////////////////
// Grycuk function  
// 
//  Purpose : Compute the R6 integral "I" of a sphere
//           I = (3/4pi) Integral (1/|r|^6 ) dVol
//           See details in the paper or tutorial
//
//  Parameters:
//    b_i : Radius of the atom "i" from which we want to comput the 
//          invers of born R6 radius
//    b_j : Radius of atom "j" 
//    r_ij: Distance between the center of atoms "i" and "j" 
//
//  Return : The r6 integral over the sphere "j"
//
//  Calls   : calculate() in Analytic.cpp 
//            Initializachunks() in Molecule.cpp
//
//  Modified by Boris Aguilar
double Functions::Grycuk(double b_i, double b_j, double r_ij) {
	
	// Compute the absolute value of (b_i - b_j)
	double Difference = b_i - b_j;
	if (Difference < 0) {Difference = -Difference;}


	// Handle Four possible configurations of two spheres	
	if (r_ij <= Difference ) {
        	/*b_i swallowed b_j, don't count anything.*/
        	if (b_i >= r_ij + b_j) { return 0.0 ; }

        	/*b_j swallowed by b_i, this should be counted. Right*/
        	/*now, however, we don't.*/
        	else if (b_j > r_ij + b_i) {
                	cerr << "Atom swalloed in Grycuk: b_i=" << b_i
                 	<< " r_ij=" << r_ij << " b_j=" << b_j << endl;
                	return 0.0;
        	}
	}
        /*Atoms are far enough apart, there isn't any overlap*/
        else if (r_ij > b_i + b_j) {
                double term = b_j / (r_ij*r_ij - b_j*b_j);
                return term*term*term;
        }
        /*Atoms are overlapping.*/
        else {
                double v0 = 16 * r_ij;
                double v1 = r_ij + 3*b_j;
                double v2 = r_ij + b_j;
                double v3 = 3*(b_j*b_j - b_i*b_i - (r_ij-b_i)*(r_ij-b_i))+2*r_ij*b_i;
                double v4 = b_i*b_i*b_i*b_i;
                return (1./v0)*(v1/(v2*v2*v2)+v3/v4);
        }
}

//////////////////////////////////////////////////////////////
// SmoothFactor
// 
//  Purpose : Compute a factor that will smooth the Volumen measure  
//           See details in the paper FACTS (Caflish paper)
//
//  Parameters:
//    R_sphere : Radius of the sphere centered at atom "i" 
//   
//    r_ij: Distance between the center of atoms "i" and "j" 
//
//  Return : A factor to be multiplied to the Volume measure 
//
//  Calls   : calculate() in Analytic.cpp 
//            
//  Modified by Boris Aguilar

double Functions::SmoothFactor(double R_sphere, double r_ij) {

        
        if (r_ij  <=  R_sphere ) 
	{
		double factor;
		factor = r_ij/R_sphere;
		return (1 - factor*factor)*(1 - factor*factor);
	}
	else { 
		return 0.0;
	}
}











/*double Functions::ThreeSphereIntersection(double R1, double R2, double R3, double D12, double D13, double D23) {
	double cost_1 = (D12*D12 + D13*D13 - D23*D23) / (2 * D12 * D23);
	double cost_2 = (D23*D23 + D12*D12 - D13*D13) / (2 * D23 * D12);
	double cost_3 = (D13*D13 + D23*D23 - D12*D12) / (2 * D13 * D23);
	
	double d12 = (R1*R1 - R2*R2 + D12*D12) / (2*D12);
	double d21 = (R2*R2 - R1*R1 + D12*D12) / (2*D12);
	double d13 = (R1*R1 - R3*R3 + D13*D13) / (2*D13);
	double d31 = (R3*R3 - R1*R1 + D13*D13) / (2*D13);
	double d23 = (R2*R2 - R3*R3 + D23*D23) / (2*D23);
	double d32 = (R3*R3 - R2*R2 + D23*D23) / (2*D23);

	double r12 = sqrt((2*(R1*R1*R2*R2 + R1*R1*D12*D12+R2*R2*D12*D12)-(R1*R1*R1*R1 + R2*R2*R2*R2 + D12*D12*D12*D12))
	 / (4 * D12*D12));
	double r13 = sqrt((2*(R1*R1*R3*R3 + R1*R1*D13*D13+R3*R3*D13*D13)-(R1*R1*R1*R1 + R3*R3*R3*R3 + D13*D13*D13*D13))
	 / (4 * D13*D13));
	double r23 = sqrt((2*(R2*R2*R3*R3 + R2*R2*D23*D23+R3*R3*D23*D23)-(R2*R2*R2*R2 + R3*R3*R3*R3 + D23*D23*D23*D23))
	 / (4 * D23*D23));
	
	double costPrime_1 = (d12*d13 - r12*r13)/(R1*R1);
	double costPrime_2 = (d23*d21 - r23*r12)/(R2*R2);
	double costPrime_3 = (d31*d32 - r13*r23)/(R3*R3);
	
	double costPrime_1 = (d12*d13 + r12*r13)/(R1*R1);
	double costPrime_2 = (d23*d21 + r23*r12)/(R2*R2);
	double costPrime_3 = (d31*d32 + r13*r23)/(R3*R3);
		
	double V1, V2, V3;
	if (cost_1 <= costPrime_1) {
		V1 = 0;
	} else if (costPrime_2 <= cost_2 && cost_2 <= costDoublePrime_2) {
		V1 = W1;
	} else if (costDoublePrime_2 <= cost_2 && d12 >= d13) {
		V1 = W12;
	} else if (costDoublePrime_2 <= cost_2 && d12 <= d13) {
		V1 = W13;
	}

	if (cost_2 <= costPrime_2) {
		V2 = 0;
	} else if (costPrime_2 <= cost_2 && cost_2 <= costDoublePrime_2) {
		V2 = W2;
	} else if (costDoublePrime_2 <= cost_2 && d23 >= d21) {
		V2 = W23;
	} else if (costDoublePrime_2 <= cost_2 && d23 <= d21) {
		V2 = W21;
	}

	if (cost_3 <= costPrime_3) {
		V3 = 0;
	} else if (costPrime_3 <= cost_3 && cost_3 <= costDoublePrime_3) {
		V3 = W3;
	} else if (costDoublePrime_3 <= cost_3 && d31 >= d32) {
		V3 = W31;
	} else if (costDoublePrime_3 <= cost_3 && d31 <= d32) {
		V3 = W32;
	}
	return 0;
}*/

/*double Functions::D_0(double R_1, double R_2, double R_w) {
	char atom_1[LINE_LENGTH];
	char atom_2[LINE_LENGTH];
	
	sprintf(atom_1, "%f", R_1);
	sprintf(atom_2, "%f", R_2);

	string key = atom_1; key += atom_2;
	map<string, double>::iterator result = d_0.find(key);
	if (result == d_0.end()) {
		cerr << "Could not find d_0 value for " << R_1 << " and " << R_2 << endl;
		return 0;
	}
	return result->second;
}

double Functions::M_0(double R_1, double R_2, double R_w) {
	char atom_1[LINE_LENGTH];
	char atom_2[LINE_LENGTH];
			
	sprintf(atom_1, "%f", R_1);
	sprintf(atom_2, "%f", R_2);

	string key = atom_1; key += atom_2;
	map<string, double>::iterator result = m_0.find(key);
	if (result == m_0.end()) {
		cerr << "Could not find m_0 value for " << R_1 << " and " << R_2 << endl;
		return 0;
	}
	return result->second;
}

bool Functions::loadD_0(string fileName) {
	ifstream in(fileName.c_str());
	if (!in) {
		cerr << "Error opening file " + fileName + " for loading D_0 values." << endl;
		return false;
	}
	double start, end, stepsize;
	in >> start; in >> end; in >> stepsize;
	
	double i,j;
	for (i = start; i <= end; i += stepsize) {
		for (j = start; j <= end; j+= stepsize) {
			double value;
			in >> value;
			char atom_1[LINE_LENGTH];
			char atom_2[LINE_LENGTH];
			
			sprintf(atom_1, "%d", i);
			sprintf(atom_2, "%d", j);

			string key = atom_1; key += atom_2;
			
			d_0.insert(make_pair(key, value));
		}
	}		
}

bool Functions::loadM_0(string fileName) {
	ifstream in(fileName.c_str());
	if (!in) {
		cerr << "Error opening file " + fileName + " for loading M_0 values." << endl;
		return false;
	}
	double start, end, stepsize;
	in >> start; in >> end; in >> stepsize;
	
	double i,j;
	for (i = start; i <= end; i += stepsize) {
		for (j = start; j <= end; j+= stepsize) {
			double value;
			in >> value;
			char atom_1[LINE_LENGTH];
			char atom_2[LINE_LENGTH];
			
			sprintf(atom_1, "%f", i);
			sprintf(atom_2, "%f", j);

			string key = atom_1; key += atom_2;
			
			m_0.insert(make_pair(key, value));
		}
	}		
}*/
