#ifndef MOLECULARINTEGRATION_H
#define MOLECULARINTEGRATION_H

/////////////////////////////////////////////////////////////////////////////
// 
// calculates the r^6 integral that gives the exact Born radii for a sphere.
// Integration is done in the most 
// straightforward way: areas of plane triangles representing the MSMS
// surface are multiplied by average (\vec r \cdot \vec n) of the 
// vetices and summed up.
//
// Command line use: bornradius <MSMS filename prefix> <file.xyzqr>
// (for example, if MSMS data are in files msmsout.face and msmsout.vert,
// the prefix is msmsout)
//
// Author: Grigori Sigalov <sigalov@vt.edu>, <greg_sigalov@yahoo.com>
// Version of 05/05/08.
//
// Modified: Richard Shadrach <rshad@vt.edu>
// Version of 08/08/08
//
/////////////////////////////////////////////////////////////////////////////

using namespace std; 

#include <stdio.h> 
#include <stdlib.h> 
#include <string.h>
#include <math.h>

#include <sstream>

#include "Molecule.h"

#define NUMBORNCOEF  1

// alpha from the same paper by Sigalov et al, J Chem Phys 2006;
// this is the ab initio value for epsin=epsout

#define ALPHA        0.571412
#define SOLVENTR     1.4

// econv2 = 331.842 as used in PEP; this is close to Intl. Steam Table value
#define ECONV2 331.842

// square of the neighbor cut-off distance
#define NEIGHBORDIST2 16.0

#define saveCorrelationMatrix 0

typedef struct {
	FILE *ffd;
	FILE *vfd;
	//molfile_graphics_t *graphics;
} msms_t;

typedef struct {
	float xmin;
	float xmax;
	float dx;
	float ymin;
	float ymax;
	float dy;
	float zmin;
	float zmax;
	float dz;
	int   nx;
	int   ny;
	int   nz;
} mesh_t;

class MolecularIntegration {
public:
	static int SingleAtomR6(Molecule &m, int centerAtom, int density, double ProbeRadius );
	
	static int EntireMoleculeR6(Molecule &m, int density, double ProbeRadius, double B=0);

	static int NumericalNeckIntegration(Molecule &m, int density, double ProbeRadius);

	static int ComputeAverageRadius(Molecule &m, int density, double ProbeRadius);

private:
	static void *open_file_read(const char *filepath, const char *filetype, int *natoms);
	static float *calculate_facetarea(int facecount, int *facet, float *vertex);
	static double calculate_reverse_alpha(int facecount, int vertexcount, int *facet, float *facetarea, 
	 float *vertex, float *normal,  Atom &a);
	static float averageBornRadius(int bad, int atomcount, float *atomCoord, float *analytRadii);
	static float calculate_totalarea(int facecount, int *facet, float *vertex);
 
};

#endif
