#include "Molecule.h"
#include "MolecularIntegration.h"

/********************************Constructors********************************/
Molecule::Molecule() {
	bounded = false;
	name="";
}

Molecule::Molecule(string fileName, bool paramXYZformat) {
	bounded = false;
	loadPQRFile(fileName, paramXYZformat);
}

Molecule::Molecule(const Molecule& m) {
	atoms = m.atoms;
	name = m.name;
	XYZformat = m.XYZformat;
	overlaps = m.overlaps;
	bounded = m.bounded;
	xMin = m.xMin;
	xMax = m.xMax;
	yMin = m.yMin;
	yMax = m.yMax;
	zMin = m.zMin;
	zMax = m.zMax;
}

Molecule::Molecule(const Atoms &v) {
	atoms = v;
	bounded = false;
	XYZformat = false;
}

/***************************Operator Overloads*******************************/
ostream& operator<<(ostream& out, const Molecule& m) {
	Atoms::const_iterator itr;
	for (itr = m.atoms.begin(); itr != m.atoms.end(); itr++) {
		out << *itr << endl;
	}
	return out;
}

bool Molecule::loadRinv(string fileName) {

	ifstream in(fileName.c_str());
	if (!in) {
		cerr << "Error opening file pqr or xyz file " << fileName << endl;
		return false;
	}

	Atoms::iterator i = atoms.begin();
        double Rinv;
        int AtomNumber;

	for (i = atoms.begin(); i != atoms.end(); i++) {

                in >> AtomNumber;
                in >> Rinv;
		i->MolecularR6Sum(Rinv);                
	}
        
	if (i != atoms.end()) {
		cerr << "Error reading " << fileName << " file for chunk "
                     << "values.  Generating instead." << endl;
		return false;
	}

	return true;
}


/******************************Loading Data**********************************/
// Reading the coordinates from the PQR files or from the XZY format
bool Molecule::loadPQRFile(string fileName, bool paramXYZformat, bool chagb_flag ) {
	name = fileName;
	XYZformat = paramXYZformat;
	ifstream in(fileName.c_str());
        double offset = 0.0;
        if (chagb_flag) offset = 0.52 ;

	if (!in) {
		cerr << "Error opening file pqr or xyz file " << fileName << endl;
		return false;
	}
	double x, y, z, q, r;
	int number;
	string type, junk;
	while (in && !in.eof())	{
		if ( XYZformat ) {
			in >> x >> y >> z;
			Atom a(x, y, z, DEFAULT_ATOM_SIZE, 0);
			atoms.push_back(a);
		}
		else {
			in >> junk;
			if (junk == "ATOM" || junk == "HETATM")	{
				in >> number >> type >> junk >> junk >> x >> y >> z >> q >> r;
				Atom a(x, y, z, r + offset , number, q, type);
				atoms.push_back(a);
			}
			else {
				/* gobble down the rest of the line */
				in.ignore(255, '\n');
			}
		}
	}
	in.close();
	return true;
}

//  Reading bond information from PSF  Files
//This method assumes that atom numbers start at 1 and not 0.  If they do
//start at 0, it will terminate prematurely (probably on the first atom).
bool Molecule::loadPSFFile(string fileName) {
	ifstream in(fileName.c_str());	
	if (!in) {
		cerr << "Error loading PSF file " << fileName << endl;
		return false;
	}
	string junk;
	/*Continuously read the file till we come to the bond section*/
	while (in && !in.eof() && junk != "!NBOND:") {
		in >> junk;
	}

	/*After NBOND comes the number of bonds.  We don't need this information.*/
	in >> junk;
	
	bool isNum = true;
	int atom1, atom2;
	
	/*atom1 and atom2 should only be zero if atoi() failed*/
	in >> junk; atom1 = atoi(junk.c_str());	
	in >> junk;	atom2 = atoi(junk.c_str());
	isNum = (atom1 != 0) && (atom2 != 0);
	
	while (in && !in.eof() && isNum) {
		if (atom1 >= atoms.size() || atom2 >= atoms.size()) {
			cerr << atom1 << " or " << atom2 << " is more than the number of atoms ("
 				<< atoms.size() << ") in the PSF file." << endl;
		}
		else {
			atoms[atom1-1].addAtomToNeighbors(atom2);
			atoms[atom2-1].addAtomToNeighbors(atom1);
		}
		in >> junk;	atom1 = atoi(junk.c_str());
		in >> junk;	atom2 = atoi(junk.c_str());
		isNum = (atom1 != 0) && (atom2 != 0);
	}
	return true;
}

//Reading Bond information from topolgy files 
//Determine the neighbor atoms (covalently bonded atoms) for each
//atom of the molecule
bool Molecule::loadPRMTOPFile(string fileName) {

 ifstream in(fileName.c_str());
 int i;
 if (!in) {
	cerr << "Error loading PRMTOP file " << fileName << endl;
	return false;
 }
 string junk;

 while (in && !in.eof()) {
	if (junk == "%FLAG") {
		in >> junk;
		if (junk == "BONDS_INC_HYDROGEN" || junk == "BONDS_WITHOUT_HYDROGEN") {
			ios::pos_type pos = in.tellg();
			int count = -1;
			//Read in the %FORMAT tag
			in >> junk;
			while (in && !in.eof() && junk != "%FLAG") {
				count++;
				in >> junk;
			}
			in.seekg(pos);
			//Read in the %FORMAT tag
			in >> junk;
			for (i = 0; i < count/3; i++) {
				int a_i, a_j;
				in >> a_i >> a_j >> junk;
				if (a_i/3 >= atoms.size() || a_j/3 >= atoms.size()) {
					cerr << "Bond between " << a_i/3+1 << " and " << a_j/3+1 
						<< " was greater than number of atoms in molecule(" 
						<< atoms.size() << ").  Ignoring bond." << endl;
					continue;
				}
				atoms[a_i/3].addAtomToNeighbors(a_j/3+1);
				atoms[a_j/3].addAtomToNeighbors(a_i/3+1);
			}
		}
	}
	else {
		in >> junk;
	}
 }
 return in;


}

// Generate a PQR File 
bool Molecule::generatePQRFile(string fileName) {
	ofstream out(fileName.c_str());
	if (!out) {
		return false;
	}
	Atoms::iterator i;
	for (i = atoms.begin(); i != atoms.end() && out; i++) {
		out << "ATOM\t" << i->Number() << "\t" << i->Type() <<
		 "\t?\t1\t" << i->X() << "\t" << i->Y() << "\t" << i->Z() <<
		 "\t" << i->Charge() << "\t" << i->Radius() << endl;
	}
	return out;
}

/*******************************Initalizers**********************************/
void Molecule::initalizeOverlaps() {
	overlaps.clear();
	Atoms::iterator i,j,k;
	for (i = atoms.begin(); i != atoms.end(); i++) {
		for (j = i+1; j != atoms.end(); j++) {			
			double R = i->Radius(); double r = j->Radius();
			double d = i->distance(*j);
			if (d < R + r) {	
				Overlap o;
				/*http://mathworld.wolfram.com/Sphere-SphereIntersection.html*/
				o.volume = (M_PI*(R+r-d)*(R+r-d)*(d*d+2*d*r-3*r*r+2*d*R+6*r*R-3*R*R)) / (12*d);
				double h_1 = d - r, h_2 = d - R;
				o.x = ((d - r)*(j->X()-i->X())/d + i->X() + (d - R)*(i->X() - j->X())/d + j->X()) / 2;
				o.y = ((d - r)*(j->Y()-i->Y())/d + i->Y() + (d - R)*(i->Y() - j->Y())/d + j->Y()) / 2;
				o.z = ((d - r)*(j->Z()-i->Z())/d + i->Z() + (d - R)*(i->Z() - j->Z())/d + j->Z()) / 2;				
				overlaps.push_back(o);
			}
		}
	}
}

/*Divide and conquer algorithm.  Divides the entire region the molecule is in
/*into blocks of a small size, then checks only the atoms in adjacent blocks.
/*This method is to be used if no PSF file is given.*/
void Molecule::initalizeNeighbors() {
	int comparisons = 0, neighbors = 0;
	Atoms::iterator i,j,k;
	/*We need the bounds of the molecule to divide the region up.*/
	initalizeBounds();
	double maxRadius = 0;
	for (i = atoms.begin(); i != atoms.end(); i++) {
		i->resetNeighbors();
		if (maxRadius < i->OrigRadius()) maxRadius = i->OrigRadius();
	}
	double blockSize = 2*maxRadius;
	
	int blocksX, blocksY, blocksZ;
	blocksX = (int) ( (xMax - xMin) / blockSize ) + 1;
	blocksY = (int) ( (yMax - yMin) / blockSize ) + 1;
	blocksZ = (int) ( (zMax - zMin) / blockSize ) + 1;
	vector<int> ***grid;

	//The following few lines are simply grid allocation.
	grid = new vector<int>**[blocksX];
	int x, y, z;
	for (x = 0; x < blocksX; x++) {
		grid[x] = new vector<int>*[blocksY];
		for (y = 0; y < blocksY; y++) {
			grid[x][y] = new vector<int>[blocksZ];
		}
	}
	
	//Now put each atom into it's grid spot.
	for (i = atoms.begin(); i != atoms.end(); i++) {
		x = (int) ( (i->X() - xMin) / blockSize );
		y = (int) ( (i->Y() - yMin) / blockSize );
		z = (int) ( (i->Z() - zMin) / blockSize );
		grid[x][y][z].push_back(i->Number());
	}
	
	for (i = atoms.begin(); i != atoms.end(); i++) {
		//"pool" is a collection of all possibly bonded atoms.
		vector<int> pool;
		int addX, addY, addZ;
		x = (int) ( (i->X() - xMin) / blockSize );
		y = (int) ( (i->Y() - yMin) / blockSize );
		z = (int) ( (i->Z() - zMin) / blockSize );

		//We check each dimenions with position -1, 0, and +1.
		//But we skip it if we happen to go over (or under) a boundary
		for (addX = -1; addX < 2; addX++) {
			if (x+addX < 0 || x+addX >= blocksX) continue; 
			for (addY = -1; addY < 2; addY++) {
				if (y+addY < 0 || y+addY >= blocksY) continue;
				for (addZ = -1; addZ < 2; addZ++) {
					if (z+addZ < 0 || z+addZ >= blocksZ) continue;
					//Place all the atoms in this grid section into the pool.
					pool.insert(pool.end(), grid[x+addX][y+addY][z+addZ].begin(), 
						grid[x+addX][y+addY][z+addZ].end());
				}
			}
		}
		const double k = 0.1;	
		vector<int>::iterator j;
		for (j = pool.begin(); j != pool.end(); j++) {
			//It makes no sense to compare 1 to 2, and then 2 to 1.
			if (i->Number() >= *j) continue;
			//Remember 1-based indexing.
			Atom *a_j = &atoms[*j-1];
			if (i->distance(*a_j) < i->OrigRadius() + a_j->OrigRadius() - k) {
				i->addAtomToNeighbors(*j);
				a_j->addAtomToNeighbors(i->Number());
			}
		}
	}
}


// Generate the 
void Molecule::initalizeChunks(string chunkFile, string topFile, int depth) {
	ifstream in(chunkFile.c_str());
	Atoms::iterator i = atoms.begin();

	if (in) {
		string topologyFile;
		int d = 0;
		//First two lines of the chunk file are the topology file used to make bonds
		//and the depth (max distance) of the chunk.  If these don't agree, we
		//recalculate the chunks.
		in >> topologyFile;
		in >> d;
		if (topFile == topologyFile && depth == d) {
			//Chunk line format is value of the chunk followed by numbers of the atoms
			//in the chunk, separated by white spaces and a trailing 0 to signify the
			//end of the line.
			for (i = atoms.begin(); i != atoms.end(); i++) {
				double value;
				in >> value;
				if (!in || in.eof()) break;
				i-> ChunkR6BR(value);
				i-> ChunkR6Packing(value);
				do {
					int atom;
					in >> atom;
					if (atom == 0) break;	
					i-> addAtomToChunk(atom);
				} while (in && !in.eof());
				do {
					int atom;
					in >> atom;
					if (atom == 0) break;
					i->addAtomToNoNeck(atom);
				} while (in && !in.eof());
			}
			if (i != atoms.end()) {
				cerr << "Error reading " << chunkFile << " file for chunk "
				 << "values.  Generating instead." << endl;
			}
		}
	}
	//If the above fails to get all (or any) atoms, then redo the work and make a
	//new file.
	if (i != atoms.end()) {
		//This should probably be removed, nice to know when you're generating chunks
		//cerr << "Generating chunks..." << endl;
		in.close();
		ofstream out(chunkFile.c_str());
		out << topFile << endl;
		out << depth << endl;
		set<int>::iterator j, k;
		int c;
		for (i = atoms.begin(); i != atoms.end(); i++) {
			if (depth == -1) {
				

				Atoms::iterator j;
				for (j = atoms.begin(); j != atoms.end(); j++) {
					i->addAtomToChunk(j->Number());
				}
			}
			else if (depth == 0) {
                                i->addAtomToChunk(i->Number());
				i->ChunkR6Packing(0.0);

			
			}
			else {
				/* This part was modified by B. Aguilar  */
				/* --------------------------------------- */
				/* there many definitios of chunk  */
				// chunktype = 3 ; New definition, just consider 
				// children of nigbors as the atoms of chunk,
				// chunktype = 4 ; Atoms with only one bond are special cases
				// SDefault chunktype = 3;

				int ChunkType = 4;


				//I'm always in my own chunk
				i->addAtomToChunk(i->Number());

				int NewDepth = depth;
				if ( ChunkType == 4 && (i->neighborsSize()) == 1  ){
					// Count the number of next neigbors
					NewDepth = NewDepth + 1;
				}
                                
                                
				//Each iteration adds all the atoms within distance 1 of the chunk.
				for (c = 0; c < NewDepth; c++) {
					//Create a copy to avoid iterator invalidation
					set<int> chunk = i->Chunk();
					set<int>::iterator itr;
					for (itr = chunk.begin(); itr != chunk.end(); itr++) {
						//Remember 1-based indexing
						set<int> neighbors = atoms[*itr-1].Neighbors();
                                                set<int>::iterator itr2;
  						for (itr2 = neighbors.begin(); itr2 != neighbors.end(); itr2++) {
							set<int> c = i->Chunk();
  							// Add atom to the chunk of i (it will be removed acordeing to the depth criteria)
							i->addAtomToChunk(*itr2);

						}
						//i->addGroupToChunk(atoms[*itr-1].Neighbors());
					}
				}
			}

			//Create a "molecular" copy for the chunk that we can pass
			if (depth > 0 ) {

				Molecule chunk;
				//VERY IMPORTANT: i must appear as the first atom in this molecule.
				//We only calculate the integral on the first atom to save time.
				chunk.addAtom(atoms[i->Number()-1]);
				for (j = i->chunkBegin(); j != i->chunkEnd(); j++) {
					if (*j == i->Number()) continue;
					//Remember 1-based indexing
					chunk.addAtom(atoms[*j - 1]);
				}
				//And then find the integral value for the chunk
				MolecularIntegration::SingleAtomR6(chunk, 0, 10, 1.4);
				//We now need to copy all the values calcuated for the chunk to the atom.
				//See Atom.h for details on these functions.
				i->ChunkR6BR(chunk.atoms[0].ChunkR6BR());


				//Compute the packing Factor of chunks
				Atom a_i = *i;
                		double b_i = a_i.OrigRadius();
				// Contribution of the chunk considered as a collection of spheres
                		double approx_sum_ch = 0.0;
                		for (j = i->chunkBegin(); j != i->chunkEnd(); j++) {
                        		Atom a_j = atoms[*j-1];
                        		double r_ij = a_i.distance(a_j);
                        		double b_j = a_j.Radius();
                        		approx_sum_ch = approx_sum_ch + Functions::Grycuk(b_i, b_j, r_ij);

                		}
  				// Numerically exact contribution of chunk
                		double r3 = i->OrigRadius()*i->OrigRadius()*i->OrigRadius();
                		double chunkBR = i->ChunkR6BR();
                		double exact_sum_ch = 1/r3 - chunkBR*chunkBR*chunkBR;
                		// Compute the packing factor fo the chunk of atom i
                		double PackingFactor = exact_sum_ch / approx_sum_ch ;
                		i->ChunkR6Packing(PackingFactor);

				//The following code has been moved to the end
				//of the method in order to output the results
				//of the nonecks as well.
				/*// Save to the file
				//And save this to the file as well.
				out << PackingFactor << " ";
				for (j = i->chunkBegin(); j != i->chunkEnd(); j++) {
					out << *j << " ";
				}
				//Append a 0 to denote the end of the line.
				out << "0" << endl;*/
			}
		}
		if (depth == -1) {
			MolecularIntegration::EntireMoleculeR6(*this, 10, 1.4,0);
			for (i = this->begin(); i != this->end(); i++) {
				i->ChunkR6BR(i->MolecularR6Sum());

  				//Compute the packing Factor of chunks		
                                Atom a_i = *i;
                                double b_i = a_i.OrigRadius();
                                // Contribution of the chunk considered as a collection of spheres
                                double approx_sum_ch = 0.0;
				set<int>::iterator j;
                                for (j = i->chunkBegin(); j != i->chunkEnd(); j++) {
                                        Atom a_j = atoms[*j-1];
                                        double r_ij = a_i.distance(a_j);
                                        double b_j = a_j.Radius();
                                        approx_sum_ch = approx_sum_ch + Functions::Grycuk(b_i, b_j, r_ij);

                                }
                                // Numerically exact contribution of chunk
                                double r3 = i->OrigRadius()*i->OrigRadius()*i->OrigRadius();
                                double chunkBR = i->ChunkR6BR();
                                double exact_sum_ch = 1/r3 - chunkBR*chunkBR*chunkBR;
                                // Compute the packing factor fo the chunk of atom i
                                double PackingFactor = exact_sum_ch / approx_sum_ch ;
                                i->ChunkR6Packing(PackingFactor);
				
				//The following code has been moved to the end
				//of the method in order to output the results
				//of the nonecks as well.
				// Save to the file
				/*out << PackingFactor << " ";
				for (j = i->chunkBegin(); j != i->chunkEnd(); j++) {
					out << *j << " ";
				}
				out << "0" << endl;*/
			}
		}

		initalizeNoNecks(chunkFile, topFile, depth);

		//Output the results to the chunk file
		for (i = this->begin(); i != this->end(); i++) {
			out << i->ChunkR6Packing() << " ";
			for (j = i->chunkBegin(); j != i->chunkEnd(); j++) {
				out << *j << " ";
			}
			out << "0 ";
			for (j = i->noneckBegin(); j != i->noneckEnd(); j++) {
				out << *j << " ";
			}
			out << "0" << endl;
		}
		out.close();
	}


}

void Molecule::computeChunksR6(string chunkFile, string pqrFile, int depth) {
	string sourceFile;
	int d = 0;

	Atoms::iterator i;
	//Atoms::iterator j;

	ifstream in(chunkFile.c_str());
	in >> sourceFile;
	in >> d;
	for (i = atoms.begin(); i != atoms.end(); i++) {
		if (!in || in.eof()) break;
		do {
			int atom;
			in >> atom;
			if (atom == 0) break;
			i->addAtomToChunk(atom);
		} while (in && !in.eof());
	}
	if (i != atoms.end()) {
		cerr << "Error reading "<<chunkFile<<" Program stoped" << endl;
	} 
	else {
		cout << pqrFile << endl;
		cout << depth << endl;
		for (i = atoms.begin(); i != atoms.end(); i++) {
			Molecule chunk;
			chunk.addAtom(atoms[i->Number()-1]);
			set<int>::iterator j;
			for (j = i->chunkBegin(); j != i->chunkEnd(); j++) {
				if (*j == i->Number()) continue;
				chunk.addAtom(atoms[*j - 1]);
			}
			MolecularIntegration::SingleAtomR6(chunk,0,10, 1.4);
			i->ChunkR6BR(chunk.atoms[0].ChunkR6BR());
			cout << i->ChunkR6BR() << " ";
			for (j = i->chunkBegin(); j != i->chunkEnd(); j++) {
				cout << *j << "-";
			}
			cout << "0" << endl;
		}
	}
//	Atoms::iterator i;
//	for (i = atoms.begin(); i != atoms.end(); i++) {
//		cout << "ATOM\t" << i->Number() << endl ;
//	}

}

void Molecule::initalizeBounds() {
	/*Multiple operations require that the atoms bounds be found.
	/*We don't need to do this over an over again, check the flag first.*/
	if (!bounded) {
		bounded = true;
		Atoms::iterator i = atoms.begin();
		xMin=xMax=i->X(); yMin=yMax=i->Y(); zMin=zMax=i->Z();
		for (i = atoms.begin(); i != atoms.end(); i++) {
			if (i->X() - i->Radius() < xMin) xMin = i->X() - i->Radius();
			if (i->X() + i->Radius() > xMax) xMax = i->X() + i->Radius();
			if (i->Y() - i->Radius() < yMin) yMin = i->Y() - i->Radius();
			if (i->Y() + i->Radius() > yMax) yMax = i->Y() + i->Radius();
			if (i->Z() - i->Radius() < zMin) zMin = i->Z() - i->Radius();
			if (i->Z() + i->Radius() > zMax) zMax = i->Z() + i->Radius();
		}
	}
}

/****************************Radii Adjustments*******************************/
void Molecule::adjustRadii(double param1, double param2, double param3, 
	double (*func)(double, double, double, double)) {
	Atoms::iterator i;
	set<int>::iterator j;
	for (i = atoms.begin(); i != atoms.end(); i++) {
		double R = i->Radius();
		double volume = 4.0/3.0 * M_PI * R*R*R;
		double overlaps = 0;
		for (j = i->neighborsBegin(); j != i->neighborsEnd(); j++) {
			Atom a_j = *(atoms.begin() + (*j-1));
			double d = i->distance(a_j);
			double r = a_j.Radius();
			if (d < R + r) {
				/*http://mathworld.wolfram.com/Sphere-SphereIntersection.html*/
				overlaps += ((M_PI*(R+r-d)*(R+r-d)*(d*d+2*d*r-3*r*r+2*d*R+6*r*R-3*R*R)) / (12*d));
			}
		}
		if (overlaps != 0) {
			volume -= func(overlaps, param1, param2, param3);
		}
		/*Converts the new volume into a radius*/
		i->Radius( Functions::CubeRoot(volume * 3 / (4.0 * M_PI)) );
	}
}
void Molecule::restoreOriginalRadii() {
	Atoms::iterator i;
	for (i = atoms.begin(); i != atoms.end(); i++) {
		i->Radius(i->OrigRadius());
	}
}

/********************************Accessors***********************************/
Atoms::const_iterator Molecule::begin() const {
	return atoms.begin();
}
Atoms::const_iterator Molecule::end() const {
	return atoms.end();
}
size_t Molecule::size() const {
	return atoms.size();
}
double Molecule::XMin() const {
	return xMin;
}
double Molecule::XMax() const {
	return xMax;
}
double Molecule::YMin() const {
	return yMin;
}
double Molecule::YMax() const {
	return yMax;
}
double Molecule::ZMin() const {
	return zMin;
}
double Molecule::ZMax() const {
	return zMax;
}


double Molecule::Xav() const {
	return x_av;
}

double Molecule::Yav() const {
	return y_av;
}

double Molecule::Zav() const {
	return z_av;
}

double Molecule::Rav() const {
	return r_av;
}
/********************************Modifiers***********************************/
Atom &Molecule::getAtom(int atomNumber) {
	/*Remember 1-based indexing*/
	return atoms[atomNumber-1];
}
void Molecule::addAtom(const Atom &a) {
	bounded = false;
	atoms.push_back(a);
}
Atoms::iterator Molecule::begin() {
	return atoms.begin();
}
Atoms::iterator Molecule::end() {
	return atoms.end();
}
vector<Overlap>::iterator Molecule::overlapsBegin() {
	return overlaps.begin();
}
vector<Overlap>::iterator Molecule::overlapsEnd() {
	return overlaps.end();
}


void Molecule::Xav(double paramX) {
	x_av = paramX ;
}

void Molecule::Yav(double paramY) {
	y_av = paramY ;
}

void Molecule::Zav(double paramZ) {
	z_av = paramZ ;
}
void Molecule::Rav(double paramR) {
	r_av = paramR ;
}


// ---- Set up atopm region
int Molecule::SetAtomRegion(int atomNumber, int NewRegion) {
        atoms[atomNumber-1].Region = NewRegion;
}


//-----------------------------Helper Functions--------------------------------
bool Molecule::depthFirstSearch(Atom target, Atom a, int maxdepth, int depth) {
  if (target == a) return true;
	if (depth == maxdepth) return false;
	//cout << "      Preforming depth first search at level " << depth << " from atom " << a.Number() << 
  // " looking for " << target.Number() << endl;
	set<int>::iterator i;
	for (i = a.neighborsBegin(); i != a.neighborsEnd(); i++) {
    //cout << "         Looking at " << a.Number() << "'s neighbor " << *i << "...";
		if (atoms[*i-1] == target) {
      //cout << "FOUND" << endl;
      return true;
    }
    //cout << "NO" << endl;
		if (depthFirstSearch(target, atoms[*i-1], maxdepth, depth+1)) {
			return true;
		}
	}
	return false;
}

void Molecule::initalizeNoNecks(string chunkFile, string topFile, int depth) {
        ifstream in(chunkFile.c_str());
        Atoms::iterator i = atoms.begin();


        for (i = atoms.begin(); i != atoms.end(); i++) {
                if (depth == -1) {
                        Atoms::iterator j;
                        for (j = atoms.begin(); j != atoms.end(); j++) {
                                i->addAtomToNoNeck(j->Number());
                        }
                }
		else if (depth == 0) {
			i->addAtomToNoNeck(i->Number());
		}
                else {
                        //I'm always in my own chunk
                        i->addAtomToNoNeck(i->Number());

                        int ChunkType = 4;
                        int NewDepth = 4;
			NewDepth = depth+1;
                        if ( ChunkType == 4 && (i->neighborsSize()) == 1  ){
                                // Count the number of next neigbors
                                NewDepth = NewDepth + 1;
                        }

                        //Each iteration adds all the atoms within distance 1 of the chunk.
                        for (int c = 0; c < NewDepth; c++) {
                                //Create a copy to avoid iterator invalidation
                                set<int> noneck = i->NoNeck();
                                set<int>::iterator itr;
                                for (itr = noneck.begin(); itr != noneck.end(); itr++) {
                                        i->addGroupToNoNeck(atoms[*itr-1].Neighbors());
                                }
                         }

                }

        }

}

void Molecule::ZAP9radii( string topFile ) {

   if (topFile == "" ) {
       cerr << "the topology file is required  ...exiting"<<endl;
       exit(1);
   }


   double rC = 1.87;
   double rH = 1.1;
   double rHN = 1.2 ; // Hydrogen in secondary N
   double rN = 1.55 ;
   double rNS = 1.4 ; // Secondary
   double rNT = 1.4 ; // Tertiary
   double rO = 1.52;
   double rOC = 1.76; // carbonyl
   double rS = 2.15;
   double rP = 1.8;
   double rF = 2.40;
   double rCl = 1.82;
   double rBr = 1.85;
   double rI = 2.65;


   //Atoms::iterator i = atoms.begin();
   //for (i = atoms.begin(); i != atoms.end(); i++) {

   Atoms::iterator j;
   for (j = atoms.begin(); j != atoms.end(); j++) {
      j -> Radius(0.0);
   }

   for (j = atoms.begin(); j != atoms.end(); j++) {

      if  (j -> Radius() != 0.0 ) continue;
      string junk  = j -> Type();
      string AtomName;

      if ( isdigit(junk[0]) )  AtomName = junk.erase(0,1);
      else AtomName = junk;

      if (AtomName[0] == 'I' ) j -> Radius(rI);
      else if (AtomName[0] == 'F' ) j -> Radius(rF) ;
      else if (AtomName[0] == 'P' ) j -> Radius(rP) ;
      else if (AtomName[0] == 'S' ) j -> Radius(rS) ;
      else if (AtomName[0] == 'O' ){
         j -> Radius (rO);
         // Set the region for chunks and NoNecks 
         if (j -> neighborsSize() == 1 ){
            set<int>::iterator itr;
            itr = j-> neighborsBegin();
            Atom Aj = atoms[*itr - 1];
            string neighbor = Aj.Type();
            if ( neighbor[0] == 'C') j -> Radius(rOC);
         }
      }
      else if (AtomName[0] == 'N' ) {
         j -> Radius(rN);
         int numberH =0;
         set<int>::iterator itr;
         if (j -> neighborsSize() < 3 ) continue;

         for ( itr = j-> neighborsBegin() ;itr != j-> neighborsEnd(); itr++) {
            Atom Aj = atoms[*itr - 1 ];
            string neighbor = Aj.Type();
            if ( neighbor[0] == 'H')  numberH = numberH + 1;
         }

         if ( numberH == 1 ) { // Secondary
            for ( itr = j-> neighborsBegin() ;itr != j-> neighborsEnd(); itr++) {
               Atom Aj = atoms[ *itr - 1 ];
               string neighbor = Aj.Type();
               if ( neighbor[0] == 'H')   atoms[*itr - 1  ].Radius(rHN) ;
            }
            j -> Radius(rNS);
         }
         else if ( numberH == 0) j -> Radius(rNT); // Tertiary
      }

      else if (AtomName[0] == 'H' ) j -> Radius(rH) ;
      else if (AtomName[0] == 'C' ) {
         j -> Radius(rC);
         if ( (AtomName.size() > 1) && (AtomName[1] == 'l'))  j -> Radius(rCl)  ;
      }
      else if (AtomName[0] == 'B' ) j -> Radius(rBr) ;

      else {
         cerr << " Atom not found  " << j -> Number();
         exit(1);
      }
   }
}

// function that change the intrinsic raddi of the current
// Molecule Object
// inputs  
// chagb_flag :  flag to find if chgb model is being used
// outputs none 
void Molecule::BestGBradii( bool chagb_flag ) {

   // Best fit for CHAGB
   double rC  = 1.60;
   double rH  = 0.52;
   double rN  = 1.58;
   double rO  = 1.36;
   double rS  = 1.72;
   double rP  = 1.63;
   double rF  = 1.22;
   double rCl = 1.63;
   double rBr = 1.84;
   double rI  = 2.14;
// Ions obtained from Marcus 1988   
   double rLi = 0.74;
   double rK  = 1.38;
   double rRb = 1.49;
   double rCs = 1.70;
   double rCM = 1.81;
   double rCa = 1.00;
   double rCu = 0.73;
   double rMg = 0.72;
   double rNa = 1.02;

   double offset = 0.52 ;

   if ( chagb_flag == false ) { // best fit for canonical GB
      rC  = 1.85;
      rH  = 1.30;
      rN  = 1.40;
      rO  = 1.49;
      rS  = 1.46;
      rP  = 1.20;
      rF  = 0.82;
      rCl = 1.87;
      rBr = 1.47;
      rI  = 1.31;
// Ions only   
      rLi = 0.74;
      rK  = 1.38;
      rRb = 1.49;
      rCs = 1.70;
      rCM = 1.81;
      rCa = 1.00;
      rCu = 0.73;
      rMg = 0.72;
      rNa = 1.02;
   }
   else {
      rC  = 1.60 + offset ;
      rH  = 0.52 + offset ;
      rN  = 1.58 + offset ;
      rO  = 1.36 + offset ;
      rS  = 1.72 + offset ;
      rP  = 1.63 + offset ;
      rF  = 1.22 + offset ;
      rCl = 1.63 + offset ;
      rBr = 1.84 + offset ;
      rI  = 2.14 + offset ;
// Ions only   
      rLi = 0.74 + offset;
      rK  = 1.38 + offset;
      rRb = 1.49 + offset;
      rCs = 1.70 + offset;
      rCM = 1.81 + offset;
      rCa = 1.00 + offset;
      rCu = 0.73 + offset;
      rMg = 0.72 + offset;
      rNa = 1.02 + offset;
  }

   Atoms::iterator j;
   for (j = atoms.begin(); j != atoms.end(); j++) {
      j -> Radius(0.0);
   }

   for (j = atoms.begin(); j != atoms.end(); j++) {

      if  (j -> Radius() != 0.0 ) continue;
      string junk  = j -> Type();
      string AtomName;

      if ( isdigit(junk[0]) )  AtomName = junk.erase(0,1);
      else AtomName = junk;

      if (AtomName[0] == 'I' ) 
      {j -> Radius(rI);
          if (AtomName.size()>1){
            if (AtomName[1]=='M') j->Radius(rCM);
            if (AtomName[1]=='P') j->Radius(rNa);
        }
      }
      else if (AtomName[0] == 'F') 
      {
          if ((AtomName.size()>1)&&((AtomName[1]='e')||(AtomName[1]='E'))) {
              cerr<<"Atom type "<<AtomName<<" has no assigned radius: Exiting"<<endl;
              exit(1);}
              else  j -> Radius(rF) ;
      }
      else if (AtomName[0] == 'P') j -> Radius(rP) ;
      else if (AtomName[0] == 'S') {
          j -> Radius(rS) ;
      if (AtomName.size() > 1) {
      if (AtomName[1]=='r') {
              cerr<<"Atom type "<<AtomName<<" has no assigned radius: Exiting"<<endl;
              exit(1);}
      }
      }
      else if (AtomName[0] == 'O' ) j -> Radius(rO);
      else if (AtomName[0] == 'N' ) {
          j -> Radius(rN) ;
      if ( (AtomName.size() > 1) && (AtomName[1] == 'a'))  
            j -> Radius(rNa)  ;
      cerr<<"the radius(from Marcus et. al. 1988) of "<<AtomName<<" OK for isolated ions, not optimized for "<< AtomName<<" as a part of a complex."<<endl;
      }
      else if ((AtomName[0] == 'M')&&((AtomName[1]=='G')||(AtomName[1]=='g'))){ 
          j ->Radius(rMg);
      cerr<<"the radius(from Marcus et. al. 1988) of "<<AtomName<<" OK for isolated ions, not optimized for "<< AtomName<<" as a part of a complex."<<endl;
      }
      else if (AtomName[0] == 'H' ) j -> Radius(rH);
      else if (AtomName[0] == 'C' ) {
         j -> Radius(rC);
         if (AtomName.size() > 1) {
               if  (AtomName[1] == 'l') j -> Radius(rCl);
               else if (AtomName[1]=='U') {
                j -> Radius(rCu);
       cerr<<"the radius(from Marcus et. al. 1988) of "<<AtomName<<": OK for isolated ions, not optimized for "<< AtomName<<" as a part of a complex."<<endl;
               }

               else if (AtomName[1]=='s') {
                   j -> Radius(rCs);
       cerr<<"the radius(from Marcus et. al. 1988) of "<<AtomName<<": OK for isolated ions, not optimized for "<< AtomName<<" as a part of a complex."<<endl;
               }
           else if (AtomName[1]=='a') {
               j -> Radius(rCa);
       cerr<<"the radius(from Marcus et. al. 1988) of "<<AtomName<<": OK for isolated ions, not optimized for "<< AtomName<<" as a part of a complex."<<endl;
           }
         }
      }
      else if (AtomName[0]=='Q'){
      if (AtomName.size() > 1) {
      if (AtomName[1]=='C') {
          j->Radius(rCs);
       cerr<<"the radius(from Marcus et. al. 1988) of "<<AtomName<<": OK for isolated ions, not optimized for "<< AtomName<<" as a part of a complex."<<endl;
      }
      else if (AtomName[1]='K') {
          j->Radius(rK);
       cerr<<"the radius(from Marcus et. al. 1988) of "<<AtomName<<": OK for isolated ions, not optimized for "<< AtomName<<" as a part of a complex."<<endl;
      }
      else if (AtomName[1]='N') {
          j->Radius(rNa);
       cerr<<"the radius(from Marcus et. al. 1988) of "<<AtomName<<": OK for isolated ions, not optimized for "<< AtomName<<" as a part of a complex."<<endl;
      }
      else if (AtomName[1]='R') {
          j->Radius(rRb);
       cerr<<"the radius(from Marcus et. al. 1988) of "<<AtomName<<": OK for isolated ions, not optimized for "<< AtomName<<" as a part of a complex."<<endl;
      }
      }
      }
      else if ((AtomName[0] == 'L')&&((AtomName[1]=='I')||(AtomName[1]=='i'))) {
          j ->Radius(rLi);
       cerr<<"the radius(from Marcus et. al. 1988) of "<<AtomName<<": OK for isolated ions, not optimized for "<< AtomName<<" as a part of a complex."<<endl;
      }
      else if (AtomName[0] == 'B' ) {
          j -> Radius(rBr) ;
      if (AtomName.size() > 1) {
      if (AtomName[1]=='a'){ 
              cerr<<"Atom type "<<AtomName<<" has no assigned radius: Exiting"<<endl;
              exit(1);}
      }
      }
      else {
         cerr << " Atom name " << AtomName ;
         cerr << " not found, names should start with: " ;
         cerr << " C(arbon), H(hydrogen), N(itrogen), O(xygen), S(ulphur),";
         cerr << " P(hosphorus), F(luorine), Cl(orine),Br(omine), I(odine), Li(thium), Na(Sodium), Mg(magnesium), Ba(rium), Ca(lcium), CU(pper)";
         cerr << " QC(Cesium Amber), QK(Potassium Amber), QN(Sodium Amber), QR(Rubidium Amber), IM(ionic Cl-), IP(ionic Na+) "<< endl; ;
         exit(1);
      }
   }
}


