/*****************************************************************************
//Molecule.h
//Created by Richard Shadrach
//2008-7-29
//
//Description: 
//Represents a molecule, or rather, collection of atoms.  Many atom specific
//properties are initalized from here, such as neighbors, chunks, and overlaps
//This is to avoid having to iterate in external code.  One important thing to
//note is that atom numbers in pqr files typically start with 1.
//However, vectors start with 0.  This is a potential pitfall and must be
//watched carefully.  Any operations done with iterators are 0 based.  All
//other operations done through this class are 1 based.  To avoid any
//confusion, it is recommended that all code be written without the use of
//iterators.  However iterators are rather useful, and most times prefered.
//
//Definitions: 
//Neighbor: Any atom that is directly bonded to *this.
//Chunk: The largest collection of atoms around *this such that every atom
//       in the chunk is within distance 2 where distance is traversing a bond.
*****************************************************************************/

#ifndef MOLECULE_H
#define MOLECULE_H

#include <iostream>
#include <vector>
#include <string>
#include <fstream>

#define DEFAULT_ATOM_SIZE 1.5
#define TOL 1.E-12

#include "Atom.h"
typedef vector<Atom> Atoms;

using namespace std;

struct Overlap {
	double x, y, z;
	double volume;
};

class Molecule {
public:
	Molecule();
	Molecule(string fileName, bool paramXYZformat = false);
	Molecule(const Molecule& m);
	Molecule(const Atoms &v);

	friend ostream& operator<<(ostream& out, const Molecule& m);

	bool loadPQRFile(string fileName, bool paramXYZformat = false, bool chagb_flag = false );
	bool loadPSFFile(string fileName);
	bool loadRinv(string fileName);
	bool loadPRMTOPFile(string fileName);
	bool generatePQRFile(string fileName);

	void initalizeNeighbors();
	void initalizeOverlaps();
	void initalizeChunks(string chunkFile, string prmtopFile, int depth);
	
	void initalizeNoNecks(string chunkFile, string prmtopFile, int depth);

	void initalizeBounds();
        void computeChunksR6(string chunkFile, string prmtopFile, int depth);

	void adjustRadii(double param1, double param2, double param3, double (*func)(double, double, double, double));
	void restoreOriginalRadii();
        void ZAP9radii(string topFile);
        void BestGBradii( bool chagb_flag );
 

	Atoms::const_iterator begin() const;
	Atoms::const_iterator end() const;
	size_t size() const;

	Atom &getAtom(int atomNumber);
	void addAtom(const Atom &a);
	Atoms::iterator begin();
	Atoms::iterator end();
	vector<Overlap>::iterator overlapsBegin();
	vector<Overlap>::iterator overlapsEnd();

	/*Bounds of the molecule*/
	double XMin() const;
	double XMax() const;
	double YMin() const;
	double YMax() const;
	double ZMin() const;
	double ZMax() const;

	// Center of MAss 
	double Xav() const;
	double Yav() const;
	double Zav() const;

	void Xav(double paramX);
	void Yav(double paramY);
	void Zav(double paramZ);

	// Radius of the sphere withe same are of the molecule
	double Rav() const;
	void Rav(double paramRadius);

	// Probe Radius for msms
	double ProbeRadius() const;
	void ProbeRadius(double paramRadius);

        int SetAtomRegion(int atomNunber, int NewRegion );


        // Parameters for the VDW part of the solvation energy
        double sigma_w;      // Lennard Jones parameters for 
        double epsilon_w;    // TIP3P water oxygen
        double d_w; 	     //

        // Solvent Accesible Area;
        double SASarea; 
	
private:
	bool depthFirstSearch(Atom target, Atom a, int maxdepth, int depth);

	Atoms atoms;
	string name;
	bool XYZformat;
	vector<Overlap> overlaps;
	double xMin, xMax, yMin, yMax, zMin, zMax;
	bool bounded;

	double x_av, y_av, z_av, r_av;  // x,y,z coordinates of the center of mass.
                                        // r_av : Average Radius of the moleculae

	double R_w; // Probe Radius for msms  


};

#endif
