
#ifndef ELSIZE_H
#define ELSIZE_H

//////////////////////////////////////////////////////////////////////////////
// 
// elsize.cc Calculates the electrostatic size, A, of a molecule using 
// atomic coordinates and radii  from its PQR (or simply xyz) file.
// Author: Grigori Sigalov <sigalov@vt.edu>, <greg_sigalov@yahoo.com>
//
// Algorithm Details are in: 
//
// "Analytical Linearized Poisson--Boltzmann Approach 
//   for Molecular Dynamics Applications", by Sigalov, Fenley and Onufriev
//  J. Chem. Phys., 124, 124902  (2006)
//
////////////////////////////////////////////////////////////////////////////

// Brief algorithm: First, the moments of inertia of the molecule are
// calculated. If the program is run without second parameter or, which is the
// same, with second parameter being "-det", the electrostatic size is found
// from the third invariant (determinant) of the tensor of inertia.  Otherwise
// some more work has to be done. The _principal_ moments of inertia are found
// by solving the cubic equation det(I - lambda E) = 0. The semiaxes of the
// ellipsoid that has same principal moments of inertia and mass as the
// original molecule are found. Then, there are two possibilities: option
// "-ell" leads to calculation of the exact electrostatic size, A_ell, of the
// effective ellipoid just found, which involves numerical summation by
// Simpson method of an elliptic intergal of the first kind (to handle the
// infinite limit, the integral is transformed to map infinity to zero). With
// option "-elf", the value A_elf, which is an approximation to A_ell
// expressed using elementary functions only, is calculated. Option "-abc"
// prints the semiaxes of the effective ellipsoid.  Option "-tab" prints all
// of the above into a table; option -hea prints table with a header.
// Finally, option -deb prints additionally some extra information. Any of
// these options, if used at all, MUST follow the input file name. 

// By default, the input file is in PQR format. If you have only PDB file
// available, convert it to a plain XYZ format and use an extra option,
// "-xyz". In this case all atomic sizes are set to the same value,
// DEFAULTATOMSIZE.

// Shortly, the following hierarchy can be built:

// 1. -det (default) gives a reasonable estimate of the electrostatic size. In
// most cases, you will want nothing more. Moreover, A_det has simple
// analytical derivatives, so if you are going to use derivatives of the
// electrostatic size, you should choose A_det. This option involves
// calculation of the moments of inertia of the molecule, which is a very
// straighforward procedure.  

// 3. -ell gives an exact solution for an ellipsoidal molecule, and involves
// numerical integration in addition to the cubic equation. Options -abc,
// -tab, -hea, and -deb include -ell and therefore take the same amount of
// calculation. 

// one limitation of this algorithm is that it can only represent as many atoms
// as the integer type holding the number of atoms field, there is such an extension
// as long long (supporting 64 bit integer and above) which is fairly popular
// and something of this nature will eventually be needed in order for this
// algorithm to scale in the long-term.  For now we will simply type it and 
// users with this compiler extension (g++ has such a thing) may enable it by
// compiling with -DLONG_LONG_SUPPORT

#ifdef LONG_LONG_SUPPORT
typedef unsigned long long atom_count;
#else
typedef unsigned long atom_count;
#endif

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <fstream>
#include <string>

#define DEFAULTATOMSIZE 1.5
#define ACCURACY 1.E-09 // convergence of Simpson's method, stopping condition  

#include <iostream>

#define TOL 1.E-12              // values < TOL are treated as zeros


using namespace std;


class elsize {
public:
 	elsize();
	static double ComputeElSize( string pqrFile);

	static atom_count GetNumberOfAtoms(const char *fileName, int XYZ_format);
	static atom_count ReadAtomicCoordinates(const char *fileName, int XYZ_format, atom_count natoms,
         double *x, double *y, double *z, double *r2, double *r3);

	static double EllipticIntegral(double a, double b, double c, double accuracy);
	static double A_elementary_functions(double a, double b, double c);

	// a couple of auxiliary functions used in numerical integration procedure
	static double function1(double a, double b, double c, double x);
	static double function2(double a, double b, double c, double x);


	//bool debug;
private:
};

#endif
